const std = @import("std");

pub const states = @import("./states.zig");

pub const system = @import("./system.zig");

pub const World = @import("./World.zig");

pub const Exclusive = struct {
	world: *World,

	pub const Param = struct {
		world: *World,
	};

	pub fn bind(context: system.BindContext) std.mem.Allocator.Error!Param {
		return .{
			.world = context.world,
		};
	}

	pub fn init(param: *Param) Exclusive {
		return .{
			.world = param.world,
		};
	}

	pub const thread_restriction = .main;
};

pub fn Params(comptime Value: type) type {
	if (!@hasDecl(Value, "Param")) {
		@compileError("System parameters must have a Params type declaration");
	}

	return struct {
		has_head: ?*Node = null,
		has_tail: ?*Node = null,

		pub const Node = struct {
			param: Value.Param,
			has_prev: ?*Node = null,
			has_next: ?*Node = null,
		};
	};
}

pub fn Read(comptime Value: type) type {
	return Shared(Value, .{
		.thread_restriction = states.thread_restriction(Value),
		.read_only = true,
	});
}

pub const ShareInfo = struct {
	thread_restriction: states.ThreadRestriction,
	read_only: bool,
};

pub fn Shared(comptime Value: type, comptime info: ShareInfo) type {
	const value_info = @typeInfo(Value);

	const Qualified = switch (value_info) {
		.Optional => @Type(.{
			.Optional = .{
				.child = .{
					.Pointer = .{
						.is_allowzero = false,
						.sentinel = null,
						.address_space = .generic,
						.is_volatile = false,
						.alignment = @alignOf(Value),
						.size = .One,
						.child = Value,
						.is_const = info.read_only,
					},
				},
			},
		}),

		else => @Type(.{
			.Pointer = .{
				.is_allowzero = false,
				.sentinel = null,
				.address_space = .generic,
				.is_volatile = false,
				.alignment = @alignOf(Value),
				.size = .One,
				.child = Value,
				.is_const = info.read_only,
			},
		}),
	};

	return struct {
		res: Qualified,

		const Self = @This();

		pub const Param = struct {
			res: Qualified,
		};

		pub fn bind(context: system.BindContext) std.mem.Allocator.Error!Param {
			const thread_restriction_name = switch (info.thread_restriction) {
				.main => "main thread-restricted ",
				.none => ""
			};

			const res = switch (info.read_only) {
				true => (try context.register_readable_state_access(Value)),
				false => (try context.register_writable_state_access(Value)),
			};

			return .{
				.res = switch (value_info) {
					.Optional => res,

					else => res orelse {
						@panic(std.fmt.comptimePrint("attempt to use {s}{s} {s} that has not yet been set", .{
							thread_restriction_name,
							if (info.read_only) "read-only" else "read-write",
							@typeName(Value),
						}));
					},
				},
			};
		}

		pub fn init(param: *Param) Self {
			return .{
				.res = param.res,
			};
		}
	};
}

pub fn Write(comptime Value: type) type {
	return Shared(Value, .{
		.thread_restriction = states.thread_restriction(Value),
		.read_only = false,
	});
}

fn parameter_type(comptime Value: type) *const system.Info.Parameter {
	const ValueParams = Params(Value);

	if (@sizeOf(Value) == 0) {
		@compileError("System parameters must have a non-zero size");
	}

	const parameters = struct {
		fn bind(allocator: std.mem.Allocator, context: system.BindContext) std.mem.Allocator.Error!*anyopaque {
			const value_name = @typeName(Value);

			if (!@hasDecl(Value, "bind")) {
				@compileError(
					"a `bind` declaration on " ++
					value_name ++
					" is requied for parameter types with a `Param` declaration");
			}

			const bind_type = @typeInfo(@TypeOf(Value.bind));

			if (bind_type != .Fn) {
				@compileError("`bind` declaration on " ++ value_name ++ " must be a fn");
			}

			if (bind_type.Fn.params.len != 1 or bind_type.Fn.params[0].type.? != system.BindContext) {
				@compileError(
					"`bind` fn on " ++
					value_name ++
					" must accept " ++
					@typeName(system.BindContext) ++
					" as it's one and only argument");
			}

			const params_node = try allocator.create(ValueParams.Node);

			params_node.* = .{
				.param = switch (bind_type.Fn.return_type.?) {
					Value.Param => Value.bind(context),
					std.mem.Allocator.Error!Value.Param => try Value.bind(context),

					else => @compileError(std.fmt.comptimePrint("`bind` fn on {s} must return {s} or {s}", .{
						@typeName(Value),
						@typeName(Value.Param),
						@typeName(std.mem.Allocator.Error!Value.Param)
					})),
				},
			};

			if (context.world.get_state(ValueParams)) |value_params| {
				if (value_params.has_tail) |tail| {
					tail.has_next = params_node;
				}

				params_node.has_prev = value_params.has_tail;
				value_params.has_tail = params_node;
			} else {
				try context.world.set_state(ValueParams{
					.has_head = params_node,
					.has_tail = params_node,
				});
			}

			return @ptrCast(params_node);
		}

		fn init(argument: *anyopaque, erased_node: *anyopaque) void {
			const value_name = @typeName(Value);

			if (!@hasDecl(Value, "init")) {
				@compileError("an `init` declaration on " ++ value_name ++ " is requied for parameter types");
			}

			const init_type = @typeInfo(@TypeOf(Value.init));

			if (init_type != .Fn) {
				@compileError("`init` declaration on " ++ value_name ++ " must be a fn");
			}

			if (init_type.Fn.return_type.? != Value) {
				@compileError("`init` fn on " ++ value_name ++ " must return a " ++ value_name);
			}

			const concrete_argument = @as(*Value, @ptrCast(@alignCast(argument)));

			if (init_type.Fn.params.len != 1 or init_type.Fn.params[0].type.? != *Value.Param) {
				@compileError("`init` fn on " ++ value_name ++ " must accept a " ++ @typeName(*Value.Param));
			}

			concrete_argument.* = Value.init(&@as(*ValueParams.Node, @ptrCast(@alignCast(erased_node))).param);
		}

		fn unbind(allocator: std.mem.Allocator, erased_node: *anyopaque, context: system.UnbindContext) void {
			if (@hasDecl(Value, "unbind")) {
				const node = @as(*ValueParams.Node, @ptrCast(@alignCast(erased_node)));

				if (node.has_prev) |prev| {
					prev.has_next = node.has_next;
				}

				if (node.has_next) |next| {
					next.has_prev = node.has_prev;
				}

				if (context.world.get_state(ValueParams)) |params| {
					if (node.has_prev == null) {
						params.has_head = node.has_next;
					}

					if (node.has_next == null) {
						params.has_tail = node.has_prev;
					}
				}

				Value.unbind(&node.param, context);
				allocator.destroy(node);
			}
		}
	};

	return comptime &.{
		.thread_restriction = if (@hasDecl(Value, "thread_restriction")) Value.thread_restriction else .none,
		.init = parameters.init,
		.bind = parameters.bind,
		.unbind = parameters.unbind,
	};
}

pub fn system_fn(comptime call: anytype) *const system.Info {
	const Call = @TypeOf(call);

	const system_info = comptime generate: {
		switch (@typeInfo(Call)) {
			.Fn => |call_fn| {
				if (call_fn.params.len > system.max_parameters) {
					@compileError("number of parameters to `call` cannot be more than 16");
				}

				const systems = struct {
					fn run(parameters: []const *const system.Info.Parameter, data: *const [system.max_parameters]*anyopaque) anyerror!void {
						var call_args = @as(std.meta.ArgsTuple(Call), undefined);

						inline for (parameters, &call_args, data[0 .. parameters.len]) |parameter, *call_arg, state| {
							parameter.init(call_arg, state);
						}

						switch (@typeInfo(call_fn.return_type.?)) {
							.Void => @call(.auto, call, call_args),
							.ErrorUnion => try @call(.auto, call, call_args),
							else => @compileError("number of parameters to `call` must return void or !void"),
						}
					}
				};

				var parameters = @as([system.max_parameters]*const system.Info.Parameter, undefined);
				var thread_restriction = states.ThreadRestriction.none;

				for (0 .. call_fn.params.len) |index| {
					const CallParam = call_fn.params[index].type.?;
					const parameter = parameter_type(CallParam);

					if (parameter.thread_restriction != .none) {
						if (thread_restriction != .none and thread_restriction != parameter.thread_restriction) {
							@compileError("a system may not have conflicting thread restrictions");
						}

						thread_restriction = parameter.thread_restriction;
					}

					parameters[index] = parameter;
				}

				break: generate &.{
					.parameters = parameters,
					.parameter_count = call_fn.params.len,
					.execute = systems.run,
					.thread_restriction = thread_restriction,
				};
			},

			else => @compileError("parameter `call` must be a function"),
		}
	};

	return system_info;
}
