pub const act = @import("./act.zig");

const coral = @import("coral");

const ext = @import("./ext.zig");

const flow = @import("flow");

pub const gfx = @import("./gfx.zig");

pub const msg = @import("./msg.zig");

const std = @import("std");

pub const App = struct {
	events: *const Events,
	target_frame_time: f64,
	elapsed_time: f64,
	is_running: bool,

	pub const Events = struct {
		load: flow.World.Event,
		pre_update: flow.World.Event,
		update: flow.World.Event,
		post_update: flow.World.Event,
		render: flow.World.Event,
		finish: flow.World.Event,
		exit: flow.World.Event,
	};

	pub fn quit(self: *App) void {
		self.is_running = false;
	}
};

pub const Flag = enum {
	dump_shader_translations,

	var launch_args = [_]?[]const u8{null} ** std.enums.values(Flag).len;

	const values = std.enums.values(Flag);
};

pub const Options = struct {
	tick_rate: u64,
	execution: Execution,
	middlewares: []const *const Setup = default_middlewares,

	pub const Execution = union (enum) {
		single_threaded,
		thread_share: f32,
	};
};

pub const Read = flow.Read;

pub const Setup = fn (*flow.World, App.Events) anyerror!void;

pub const World = flow.World;

pub const Write = flow.Write;

pub const default_middlewares = &.{
	gfx.setup,
	act.setup,
};

pub fn launch_arg(flag: Flag) ?[]const u8 {
	return Flag.launch_args[@intFromEnum(flag)];
}

pub fn start(comptime setup: Setup, options: Options) fn () anyerror!void {
	const Start = struct {
		fn main() anyerror!void {
			defer {
				coral.heap.trace_leaks();
				ext.SDL_Quit();
			}

			parse_args: for (std.os.argv[1 ..]) |arg| {
				const arg_span = std.mem.span(arg);
				const arg_split_index = std.mem.indexOfScalar(u8, arg_span, '=') orelse arg_span.len;
				const arg_name = arg_span[0 .. arg_split_index];

				for (Flag.values) |value| {
					const name = @tagName(value);

					if (!std.mem.eql(u8, arg_name, name)) {
						continue;
					}

					Flag.launch_args[@intFromEnum(value)] =
						if (arg_split_index == arg_span.len)
							name
						else
							arg_span[arg_split_index ..];

					continue: parse_args;
				}
			}

			var world = try switch (options.execution) {
				.single_threaded => flow.World.init(0),

				.thread_share => |thread_share| init: {
					const cpu_count = @as(u32, @intCast(std.math.clamp(std.Thread.getCpuCount() catch |cpu_count_error| {
						@panic(switch (cpu_count_error) {
							error.PermissionDenied => "permission denied retrieving CPU count",
							error.SystemResources => "system resources are preventing retrieval of the CPU count",
							error.Unexpected => "unexpected failure retrieving CPU count",
						});
					}, 0, std.math.maxInt(u32))));

					break: init flow.World.init(coral.scalars.fractional(cpu_count, thread_share) orelse 0);
				},
			};

			defer world.deinit();

			const events = App.Events{
				.load = try world.create_event("load"),
				.pre_update = try world.create_event("pre-update"),
				.update = try world.create_event("update"),
				.post_update = try world.create_event("post-update"),
				.render = try world.create_event("render"),
				.finish = try world.create_event("finish"),
				.exit = try world.create_event("exit"),
			};

			const app = try world.set_get_state(App{
				.events = &events,
				.target_frame_time = 1.0 / @as(f64, @floatFromInt(options.tick_rate)),
				.elapsed_time = 0,
				.is_running = true,
			});

			for (options.middlewares) |setup_middleware| {
				try setup_middleware(&world, events);
			}

			try setup(&world, events);
			try world.run_event(events.load);

			const ticks_initial = std.time.milliTimestamp();
			var ticks_previous = ticks_initial;
			var accumulated_time = @as(f64, 0);

			while (app.is_running) {
				const ticks_current = std.time.milliTimestamp();
				const milliseconds_per_second = 1000.0;
				const delta_time = @as(f64, @floatFromInt(ticks_current - ticks_previous)) / milliseconds_per_second;

				app.elapsed_time = @as(f64, @floatFromInt(ticks_current - ticks_initial)) / milliseconds_per_second;
				ticks_previous = ticks_current;
				accumulated_time += delta_time;

				try world.run_event(events.pre_update);

				while (accumulated_time >= app.target_frame_time) : (accumulated_time -= app.target_frame_time) {
					try world.run_event(events.update);
				}

				try world.run_event(events.post_update);
				try world.run_event(events.render);
				try world.run_event(events.finish);
			}

			try world.run_event(events.exit);
		}
	};

	return Start.main;
}

pub const system_fn = flow.system_fn;
